# Use %matplotlib notebook instead of %matplotlib inline to get embedded interactive figures in the IPython notebook.  
# This requires recent versions of matplotlib (1.4+) and IPython (3.0+)
# Need to restart notebook or kernel and call %matplotlib notebook twice

import numpy as np
import matplotlib.pyplot as plt
from matplotlib import cm
from mpl_toolkits.mplot3d import Axes3D

# column count of X should >= 3
def plotComponents3D(X, y, labels):

    if X.shape[1] < 3:
        print('ERROR: X MUST HAVE AT LEAST 3 FEATURES/COLUMNS! SKIPPING plotComponents3D().')
        return

    fig = plt.figure()
    ax = Axes3D(fig)
    # ax = fig.add_subplot(111, projection='3d')

    colors = ['gray', 'red', 'blue', 'black','orange','green','cyan','purple']
    
    for idx, l in enumerate(labels):
        indices = np.where(y == l)[0]
        ax.plot(X[indices, 0],
                X[indices, 1],
                X[indices, 2],
                'o', 
                label= "C" + str(l), 
                c=colors[idx])

    ax.legend()


import numpy as np
import matplotlib.pyplot as plt
from .unsupervised_dimension_reductions import *

# plot the importance for all features
def plot_feature_importance(feature_importances, title, row_size = 100, figsize=(12,2)):
    
    feature_importances = np.array(feature_importances)

    # matrix chart
    ROW_SIZE = row_size # math.ceil(feature_importances.size / 100) * 10     
    rows = int((feature_importances.size - 1)/ROW_SIZE) + 1
    fig = plt.figure(figsize=(12,12*rows/ROW_SIZE+0.5))
    if title:
        plt.title(title + "\n" + 'importance marked by color depth')
    plt.axis('off')
    
    for i in range(rows):
        ax = fig.add_subplot(rows, 1, i + 1)
        ax.axis('off')
        arr = feature_importances[i*ROW_SIZE: min(i*ROW_SIZE+ROW_SIZE - 1, feature_importances.size - 1)]
        s = ax.matshow(arr.reshape(1,-1), cmap=plt.cm.Blues)    
    

    # bar chart
    plt.figure(figsize=figsize)
    if title:
        plt.title(title + "\n" + 'importance marked by bar height')
    plt.bar(range(feature_importances.size), feature_importances, alpha=1.0, width=10)
    plt.xticks([])
    plt.yticks([])
    plt.show()

'''
# plot the importance of the selected features
def plot_important_features(feature_names, feature_importances, xlabel = '', ylabel = '', title = '', figsize=(20,3)):

    assert len(feature_names) == len(feature_importances)
    plt.figure(figsize=figsize)
    plt.bar(feature_names.astype(str), feature_importances)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.xticks(rotation = 90)
    plt.title(title)
    plt.show()
'''


def visualize_important_features(X, y, coef, X_names = None, title = None, row_size = 100, eps = 1e-10):
    '''
    epsilon - cut threshold
    '''
    N = np.count_nonzero(np.abs(coef) > eps)
    biggest_fs = np.argsort(np.abs(coef))[::-1][:N] # take last N item indices and reverse (ord desc)
    X_fs = X[:,biggest_fs] # 前N个系数 non-zero

    print('Important feature Number:',N)
    print('Important feature Indice:',biggest_fs)
    if (X_names and len(X_names) == X.shape[1] ):
        print('Important features:', np.array(X_names)[biggest_fs])
    print('Important feature coefficents:', coef[biggest_fs])
    
    plot_feature_importance(np.abs(coef), title, row_size = row_size)
    unsupervised_dimension_reductions(X_fs, y)


def get_important_features(X, coef, eps = 1e-10):

    N = np.count_nonzero(np.abs(coef) > eps)
    biggest_fs = np.argsort(np.abs(coef))[::-1][:N]
    X_fs = X[:,biggest_fs]

    return X_fs


import matplotlib.pyplot as plt
import numpy as np

def plotComponents2D(X, y, labels = None, use_markers = False, ax=None, legends = None, tags = None):
    
    if X.shape[1] < 2:
        print('ERROR: X MUST HAVE AT LEAST 2 FEATURES/COLUMNS! SKIPPING plotComponents2D().')
        return
    
    # Gray shades can be given as a string encoding a float in the 0-1 range
    colors = ['0.9', '0.1', 'red', 'blue', 'black','orange','green','cyan','purple','gray']
    markers = ['o', 's', '^', 'D', 'H', 'o', 's', '^', 'D', 'H', 'o', 's', '^', 'D', 'H', 'o', 's', '^', 'D', 'H']

    if (ax is None):
        fig, ax = plt.subplots()
        
    i=0
    if (labels is None):
        labels = set(y)
        
    for label in labels:
        cluster = X[np.where(y == label)]
        # print(cluster.shape)

        if use_markers:
            ax.scatter([cluster[:,0]], [cluster[:,1]], 
                       s=40, 
                       marker=markers[i], 
                       facecolors='none', 
                       edgecolors=colors[i+3],
                       label= (str(legends[i]) if legends is not None else ("Y = " + str(label)  + ' (' + str(len(cluster)) + ')')) )
        else:
            ax.scatter([cluster[:,0]], [cluster[:,1]], 
                       s=70, 
                       facecolors=colors[i],  
                       label= (str(legends[i]) if legends is not None else ("Y = " + str(label) + ' (' + str(len(cluster)) + ')')), 
                       edgecolors = 'black', 
                       alpha = .4) # cmap='tab20'                
        i=i+1
    
    if (tags is not None):
        for j,tag in enumerate(tags):
            ax.annotate(str(tag), (X[j,0] + 0.1, X[j,1] - 0.1))
        
    ax.legend()

    ax.axes.xaxis.set_visible(False) 
    ax.axes.yaxis.set_visible(False)
    
    return ax


import matplotlib.pyplot as plt
import numpy as np

def plotComponents1D(X, y, labels, use_markers=False, ax=None, legends = None):
    
    if X is None or X.shape[1] < 1:
        print('ERROR: X Has no FEATURE/COLUMN! SKIPPING plotComponents1D().')
        return
    
    colors = ['0.8', '0.1', 'red', 'blue', 'black','orange','green','cyan','purple','gray']
    markers = ['o', 's', '^', 'D', 'H', 'o', 's', '^', 'D', 'H', 'o', 's', '^', 'D', 'H', 'o', 's', '^', 'D', 'H']

    if (ax is None):
        fig, ax = plt.subplots()
        
    i=0
    for label in labels:
        cluster = X[np.where(y == label)]
        # print(cluster.shape)

        if use_markers:
              ax.scatter([cluster[:,0]], np.zeros_like([cluster[:,0]]), 
              s=40, marker=markers[i], 
              acecolors='none', 
              edgecolors=colors[i], 
              label= (str(legends[i]) if legends is not None else ("Y = " + str(label) + ' (' + str(len(cluster)) + ')')),
              )
        else: 
              ax.scatter([cluster[:,0]], np.zeros_like([cluster[:,0]]), 
              s=70, facecolors=colors[i],   
              label= (str(legends[i]) if legends is not None else ("Y = " + str(label) + ' (' + str(len(cluster)) + ')')),
              alpha = .4)
        i=i+1

    ax.legend()


import numpy as np
import matplotlib.pyplot as plt
from sklearn.decomposition import PCA
from sklearn.decomposition import KernelPCA
from sklearn.decomposition import TruncatedSVD
from sklearn.manifold import MDS
from sklearn.manifold import TSNE
from .plotComponents2D import plotComponents2D
from .plotComponents1D import plotComponents1D


def unsupervised_dimension_reductions(X, y, labels = None):

    # %matplotlib notebook
    
    if X is None or X.shape[1] < 1:
        print('ERROR: X HAS NO FEATURE/COLUMN!')
        return
        
    if labels == None:
        labels = list(set(y))

    fig = plt.figure(figsize=(18, 10))
    # plt.title("", fontsize=14)

    ax = fig.add_subplot(331) # , projection='polar' , projection='3d'
    
    if X.shape[1] == 1:
        plotComponents1D(X, y, labels, ax = ax)
        ax.set_title('X has only 1 FEATURE/COLUMN. Plot X directly.')
        return X
        
    # Standard PCA: equals to the linear-kernel PCA
    pca = PCA(n_components=2) # keep the first 2 components
    X_pca = pca.fit_transform(X)
    plotComponents2D(X_pca, y, labels, ax = ax)
    ax.set_title ('PCA(var% ' + str(np.round(pca.explained_variance_ratio_[0:5],3)) + ')')
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)


    ax = fig.add_subplot(332)  
    X_kpca = None
    try:
        kpca = KernelPCA(n_components=2, kernel='rbf') # keep the first 2 components. default gamma = 1/n_features
        X_kpca = kpca.fit_transform(X)    
        plotComponents2D(X_kpca, y, labels, ax=ax)     
    except Exception as e:
        # if kernel PCA throws exception, print the error message
        ax.text(0.01,0.5,'rbf-kernel PCA exception: ' + str( getattr(e, 'message', repr(e) ) ) )    
    ax.set_title ('Kernel PCA (rbf)')    
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)


    ax = fig.add_subplot(333)  
    X_kpca = None
    try:
        kpca = KernelPCA(n_components=2, kernel='poly')
        X_kpca = kpca.fit_transform(X)    
        plotComponents2D(X_kpca, y, labels, ax=ax)     
    except Exception as e:
        # if kernel PCA throws exception, print the error message
        ax.text(0.01,0.5,'poly-kernel PCA exception: ' + str( getattr(e, 'message', repr(e) ) ) )    
    ax.set_title ('Kernel PCA (poly)')    
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)
    
    
    ax = fig.add_subplot(334)  
    X_kpca = None
    try:
        kpca = KernelPCA(n_components=2, kernel='cosine')
        X_kpca = kpca.fit_transform(X)    
        plotComponents2D(X_kpca, y, labels, ax=ax)     
    except Exception as e:
        # if kernel PCA throws exception, print the error message
        ax.text(0.01,0.5,'cosine-kernel PCA exception: ' + str( getattr(e, 'message', repr(e) ) ) )    
    ax.set_title ('Kernel PCA (cosine)')    
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)


    ax = fig.add_subplot(335)
    X_kpca = None
    try:
        kpca = KernelPCA(n_components=2, kernel='sigmoid')
        X_kpca = kpca.fit_transform(X)    
        plotComponents2D(X_kpca, y, labels, ax=ax)     
    except Exception as e:
        # if kernel PCA throws exception, print the error message
        ax.text(0.01,0.5,'sigmoid-kernel PCA exception: ' + str( getattr(e, 'message', repr(e) ) ) )    
    ax.set_title ('Kernel PCA (sigmoid)')    
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)


    ax = fig.add_subplot(336)
    msg_duplicate2 = ''
    if X.shape[1] == 2: # requires > n_components
        # X = np.hstack((X,X[:,-2:]))
        X_tsvd = X
        msg_duplicate2 = '\nTruncatedSVD requires X have more columns than n_components. \nDrawing the original X.'
    else:
        tsvd = TruncatedSVD(n_components=2)
        X_tsvd = tsvd.fit_transform(X)
        plotComponents2D(X_tsvd, y, labels, ax=ax)
    ax.set_title ('Truncated SVD' + msg_duplicate2 ) # PCA on uncentered data
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)
    
    
    ax = fig.add_subplot(337)
    mds = MDS(n_components=2)
    X_mds = mds.fit_transform(X)
    plotComponents2D(X_mds, y, labels, ax = ax)
    ax.set_title ('MDS')
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)


    ax = fig.add_subplot(338)    
    tsne = TSNE(n_components=2)
    X_tsne = tsne.fit_transform(X)
    plotComponents2D(X_tsne, y, labels, ax=ax)
    ax.set_title ('t-SNE')
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)

    plt.show()     

    # return X_pca, X_kpca x 4, X_tsvd, X_mds, X_tsne